<?php
/**
 * @file
 *
 * Cache class for gathering heuristics.
 */

class HeuristicCache implements DrupalCacheInterface {
  static private $_prefetch = array();
  private $prefetch = array();
  private $backend = NULL;
  protected $bin;
  private $useStaticCaching;
  private $logging = FALSE;

  function clear($cid = NULL, $wildcard = FALSE) {
    $this->log("Clearing: $cid");
    $this->invalidatePrefetch($cid, $wildcard);
    return $this->backend->clear($cid, $wildcard);
  }

  function isEmpty() {
    return $this->backend->isEmpty();
  }

  function set($cid, $data, $expire = CACHE_PERMANENT) {
    $this->log("Setting: $cid");
    if ($this->useStaticCaching) {
      $this->prefetch[$cid] = (object) array(
        'created' => REQUEST_TIME,
        'data' => $data,
      );
    }
    else {
      $this->invalidatePrefetch($cid);
    }
    return $this->backend->set($cid, $data, $expire);
  }

  function get($cid) {
    $cids = array($cid);
    $cache = $this->getMultiple($cids);
    return reset($cache);
  }

  /**
   * Constructor.
   * Initialize bin and prefetch data
   */
  function __construct($bin) {
    $this->bin = $bin;
    // Initialize prefetch for this bin
    if (!isset(self::$_prefetch[$bin])) {
      self::$_prefetch[$bin] = array();
    }
    $this->prefetch = &self::$_prefetch[$bin];

    // Let our parent do his work
    $class = $this->variable_get('class', 'DrupalDatabaseCache');
    $this->useStaticCaching = $this->variable_get('static', FALSE);
    $this->logging = $this->variable_get('logging', FALSE);

    $this->backend = new $class($bin);

    // Fetch all the cache now
    $this->prefetchCache();
  }

  /**
   * Get a variable for specific bin, fallback to default.
   */
  function variable_get($name, $default = NULL) {
    $data = variable_get('heuristic_cache_' . $name . '_' . $this->bin);
    if (!isset($data)) {
      $data = variable_get('heuristic_cache_default_' . $name, $default);
    }
    return $data;
  }

  /**
   * Aggregate cache data from prefetch and requested
   */
  function getMultiple(&$cids) {
    $fetch = array();
    $result = array();

    // Use prefetch data where available
    foreach ($cids as $cid) {
      if (
        !empty($this->prefetch[$cid]) &&
        empty($this->prefetch[$cid]->invalidated) &&
        ($this->useStaticCaching || empty($this->prefetch[$cid]->used))
      ) {
        $result[$cid] = $this->prefetch[$cid];
        $this->prefetch[$cid]->used = TRUE;
      }
      else {
        $fetch[] = $cid;
      }
    }
    $this->log("Served: " . join(', ', array_keys($result)));

    // Get all data that wasn't prefetched, from caching backend
    if ($fetch) {
      $this->log("Fetching: " . join($fetch));
      $fetched = $this->backend->getMultiple($fetch);
      $cids = $fetch;
      foreach (array_keys($fetched) as $cid) {
        $this->prefetch[$cid] = $fetched[$cid];
        $this->prefetch[$cid]->new = TRUE;
        $result[$cid] = &$fetched[$cid];
      }
    }
    else {
      $cids = array();
    }

    return $result;
  }

  /**
   * Prefetch cache
   */
  function prefetchCache() {
    $cid = 'hc:' . request_uri();
    $this->log("Prefetching $cid");
    $cids = array($cid);
    $cache = $this->backend->getMultiple($cids);
    if (isset($cache[$cid]) && is_array($cache[$cid]->data)) {
      $this->prefetch = $this->backend->getMultiple($cache[$cid]->data);
      $this->log("Missed " . join($cache[$cid]->data));
    }
    else {
      $this->prefetch = array();
    }

    // Save keys to prefetch next time.
    // Must use shutdown function. Destructor too late.
    drupal_register_shutdown_function(array($this, 'saveHeuristics'));
  }

  function invalidatePrefetch($cid, $wildcard = FALSE) {
    if ($wildcard) {
      if ($cid == '*') {
        $this->log("Invalidating: *");
        foreach ($this->prefetch as &$data) {
          $data->data = NULL;
          $data->invalidated = TRUE;
        }
      }
      else {
        foreach ($this->prefetch as $prefetch_cid => &$data) {
          if (strpos($prefetch_cid, $cid) === 0) {
            $this->log("Invalidating: $prefetch_cid");
            $data->data = NULL;
            $data->invalidated = TRUE;
          }
        }
      }
    }
    else {
      $cids = is_array($cid) ? $cid : array($cid);
      $this->log("Invalidating: " . join(',', $cids));

      // Invalidate and release memory.
      foreach ($cids as $cid) {
        if (!empty($this->prefetch[$cid])) {
          $this->prefetch[$cid]->data = NULL;
          $this->prefetch[$cid]->invalidated = TRUE;
        }
      }
    }
  }

  /**
   * Save keys to prefetch for this request (wrapper)
   */
  function saveHeuristics() {
    $this->log("Saving heuristics (wrapper)");
    drupal_register_shutdown_function(array($this, 'saveHeuristicsReal'));
  }

  /**
   * Save keys to prefetch for this request
   */
  function saveHeuristicsReal() {
    $this->log("Saving heuristics");
    $update_necessary = FALSE;
    $cids = array();
    foreach ($this->prefetch as $cid => $cache) {
      if (!empty($cache->used) || !empty($cache->new)) {
        $cids[] = $cid;
      }
      if (!empty($cache->new)) {
        $update_necessary = TRUE;
      }
    }
    if ($update_necessary && count($cids) > 1) {
      $cid = 'hc:' . request_uri();
      $this->log("Saving: " . join(', ', $cids));
      $this->backend->set($cid, $cids);
    }
  }

  function log($msg) {
    if ($this->logging) {
      error_log("[$this->bin] $msg");
    }
  }

}

